/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"

// create cnt dummy objects

void CreateDummy (int cnt)
{
    // prepare to create some dummy objects

    ooHandle(ooFDObj) federatedDBH;
    ooHandle(ooDBObj) databaseH;
    ooTrans xact;

    xact.start();
    if (!((federatedDBH.open("OO7", oocUpdate) == oocSuccess) &&
         (databaseH.open(federatedDBH, "OO7_DB", oocUpdate) == oocSuccess))) {
        fprintf(stderr, "ERROR: Cannot get hands on OO7 Database.\n");
        xact.abort();
        exit(1);
    }

    // create cnt dummy objects (in their own container)

     int DummyBytes = int (1.1 * sizeof(DummyObj) * cnt);
     int DummyPages = 1 + (DummyBytes / PageSize);
     ooHandle(ooContObj) contH = new("DummyContainer", 0, DummyPages,
       				     0, databaseH) ooContObj;
     if (contH.isNull()) {
         fprintf(stderr, "ERROR: Cannot create dummy container w/%d pages.\n",
   		 DummyPages );
        xact.abort();
        exit(1);
     }

    for (int i = 0; i < cnt; i++) {
  	ooHandle(DummyObj) dummyH = new (contH) DummyObj;
	if ((i % 1000) == 0) {
            xact.commitAndHold();
            printf("starting a new xact (i=%d)\n", i);
	}
    }
    xact.commit();
}

// scan DummySet collection retrieving all the dummy objects

void Purge()
{
     // prepare to scan all the dummy objects

     ooHandle(ooFDObj)   federatedDBH;
     ooHandle(ooDBObj)   databaseH;
     ooHandle(ooContObj) contH;
     ooTrans xact;

     xact.start();
     if (!((federatedDBH.open("OO7", oocRead) == oocSuccess) &&
          (databaseH.open(federatedDBH, "OO7_DB", oocRead) == oocSuccess))) {
         fprintf(stderr, "ERROR: Cannot get hands on OO7 Database.\n");
         xact.abort();
         exit(1);
     }
     contH.open(databaseH, "DummyContainer", oocRead);
     if (contH.isValid() == oocFalse) {
         fprintf(stderr, "ERROR: Cannot open dummy container.\n");
         xact.abort();
         exit(1);
     }

     // now do the scan to purge any/all caches

     int cnt=0;
     ooItr(DummyObj) dummyI;
     for (dummyI.scan(contH, oocRead); dummyI.next(); ) {
	cnt++;
        char c = dummyI->dummy[0]; // extract the first character
	if ((cnt%1000) == 0) {
            printf("still purging (cnt=%d)\n", cnt);
	}
     }
     xact.commit();
     printf("number of dummy objects examined = %d\n",cnt);
}
