/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"

//////////////////////////////////////////////////////////////////
//
// Insert - Create NumNewCompParts new composite parts (which
// means also creating a set of AtomicParts for each.)  Add
// a reference to each new composite part from a randomly chosen
// base assembly (a different random choice of base assembly for
// each new composite part.)
//
//////////////////////////////////////////////////////////////////


extern ooTrans	xact;
extern int	debugMode;
extern int	nextAtomicId;
extern int	nextCompositeId;
extern int	nextBaseAssemblyId;

extern ooHandle(ooContObj) DocContainerH;


//////////////////////////////////////////////////////////////////
//
// Routine to do Insert
//
//////////////////////////////////////////////////////////////////


void insert1(ooHandle(ooDBObj)& databaseH)
{
    if (debugMode) {
        printf("In Insert, nextCompositeId = %d:\n", nextCompositeId);
        printf("In Insert, nextAtomicId = %d:\n", nextAtomicId);
        printf("In Insert, nextBaseAssemblyId = %d:\n", nextBaseAssemblyId);
    }

    // get access to the design library holding composite parts

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocRead);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }
    DocContainerH.open(databaseH, "DocContainer", oocUpdate);

    // now create the desired number of new composite parts, adding each
    // one as a (private) composite part that a randomly selected base
    // assembly now wishes to use

    ooHandle(CompositePart) compH;

    int compId;
    ooKey compKey;
    compKey.type = oocInt32;
    compKey.offset = ooGetMemberOffset(CompositePart, id);
    compKey.size   = ooGetMemberSize(CompositePart, id);
    compKey.value  = &compId;

    int assmId;
    ooKey assmKey;
    assmKey.type = oocInt32;
    assmKey.offset = ooGetMemberOffset(BaseAssembly, id);
    assmKey.size   = ooGetMemberSize(BaseAssembly, id);
    assmKey.value  = &assmId;

    for (int i = 0; i < NumNewCompParts; i++) {

	// add a new composite part to the design library

        if (debugMode) {
	    printf("In Insert, making composite part %d:\n", nextCompositeId);
	}

        compId = nextCompositeId++;

        // decide if this composite part is "old" or "young" and

        compH = ooNewKey(CompositePart, (compId), designLibH, compKey);

	// randomly select a base assembly that should use it

	assmId = (int) (random() % (nextBaseAssemblyId-1)) + 1;

	// gain access to the containing module. Due to index per container
	// restriction of Objectivity/DB, by the way, need to iterate
        // through modules until we find the correct one.

        int moduleId; 
	char moduleName[40];
	ooHandle(Module) moduleH;
        ooHandle(BaseAssembly) assmH;

        for (moduleId = 1; moduleId <= TotalModules; moduleId++) {
	        sprintf(moduleName, "Module %08d", moduleId);
		moduleH.open(databaseH, moduleName, oocUpdate);
		if (moduleH.isValid() == oocFalse) {
		    fprintf(stderr, "ERROR: Unable to access %s.\n", moduleName);
		    xact.abort();
		    exit(1);
		}

		// now look for the appropriate base assembly object within its module


	        assmH.lookupObj(moduleH, assmKey);
	        if (assmH.isValid() == oocTrue) {
	 		break;   // found it!
	        } else if (moduleId == TotalModules) {
	            fprintf(stderr, "ERROR: Can't find base assembly %d in module 1\n",
		            assmId);
		    xact.abort();
	            exit(1);
	        }
        }      // if we got here must have found it!

        // finally, add the newly created composite part as a privately used
        // member of the randomly selected base assembly

	assmH->add_componentsPriv(compH);

        if (debugMode) {
	    printf("[just made it be used by base assembly %d]\n", assmId);
        }
    }
}

