/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include <ooIndex.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"

//////////////////////////////////////////////////////////////////
//
// Query #4 - randomly choose "Query4RepeatCnt" document objects by 
// lookup on their title field.  An index can be used for the
// actual lookup.  
//
/////////////////////////////////////////////////////////////////



extern ooTrans	xact;
extern int	debugMode;
extern ooHandle(ooContObj) DocContainerH;


//////////////////////////////////////////////////////////////////
//
// Routine to do Query #4
//
//////////////////////////////////////////////////////////////////


int  query4(ooHandle(ooDBObj)& databaseH)
{

    // get access to the design library containing parts and documents

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocRead);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }
    DocContainerH.open(databaseH, "DocContainer", oocRead);

    // prepare to select random documents via their title index


    ooKey key;
    key.type = oocString;
    key.offset = ooGetMemberOffset(Document, title);
    key.size   = ooGetMemberSize(Document, title);
    char title[TitleSize];
    key.value  = title;
    ooHandle(Document) docH;

    // now randomly select a document and trace a path to the base
    // assemblies that use its associated composite part (and then
    // repeat this process the desired number of times)

    // set random seed so "hot" runs are truly hot
    srandom(1);

    int count = 0;

    for (int i = 0; i < Query4RepeatCnt; i++) {

	// generate random document title and lookup document

	int compPartId = (int) (random() % TotalCompParts) + 1;
        sprintf(title, "Composite Part %08d", compPartId);
	docH.lookupObj(DocContainerH, key);
        if (docH.isValid() == oocFalse) {
            fprintf(stderr, "ERROR: Unable to find document called \"%s\"\n",
	            title);
   	    xact.abort();
            exit(1);
        }

        if (debugMode) {
            printf("    In Query4, document title = \"%s\"\n", title);
        }

	// now walk back up the path(s) to the associated base assemblies
	// (based on private uses of the composite part, at least for now)

        ooItr(BaseAssembly) baseI;
	for (docH->part()->usedInPriv(baseI); baseI.next(); ) {
	    baseI->DoNothing();
	    count++;
	}
    }
    return count;
}

