/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include <ooIndex.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"

//////////////////////////////////////////////////////////////////
//
// Query #8 --- value join between Documents and AtomicParts
// on Document.id and AtomicPart.docId.  This is a key - foreign
// key join.
//
// The main drawback is that we have no semantics whatsoever for
// this query.
//
//////////////////////////////////////////////////////////////////


extern ooTrans	xact;
extern int	debugMode;
extern ooHandle(ooContObj) DocContainerH;

extern void JoinDoNothing(int x, int y);


//////////////////////////////////////////////////////////////////
//
// Routine to do Query #8
//
//////////////////////////////////////////////////////////////////


int query8(ooHandle(ooDBObj)& databaseH)
{

    // get access to the design library containing atomic parts

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocRead);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }
    DocContainerH.open(databaseH, "DocContainer", oocRead);

    int count = 0;

    ooItr(AtomicPart) atomicPartI;

    int aid;
    ooLookupKey lookupKey(ooTypeN(Document), 1);
    ooEqualLookupField idLookupField(ooTypeN(Document), "id", &aid);
    lookupKey.addField(idLookupField);

    for (atomicPartI.scan(designLibH); atomicPartI.next(); ) {

        if (debugMode) {
            printf("    In Query8, processing atomic part %d.\n",
	                atomicPartI->id);
        }


        ooItr(Document) documentI;
        aid = atomicPartI->id;
	documentI.scan(DocContainerH, lookupKey);
	while (documentI.next()) {
	    if (debugMode) {
        	printf("In Query8, atomic part %d matches doc %d.\n",
			atomicPartI->id, documentI->id);
            }
	    JoinDoNothing(atomicPartI->id, documentI->id);
	    count++;
	}
    }
    return count;
}

